@@ -17,7 +17,7 @@ use Graph::Easy::Node::Anon;
 use Graph 0.65;
 use Graph::Directed;
 
-$VERSION = '0.19';
+$VERSION = '0.20';
 
 use strict;
 
@@ -55,7 +55,10 @@ sub _init
 
   $self->{att} = {
   node => {
-    border => '1px solid black',
+    'border' => '1px solid black',
+    'border-style' => 'solid',
+    'border-width' => '1',
+    'border-color' => 'black',
     background => 'white',
     padding => '0.2em',
     'padding-left' => '0.3em',
@@ -85,7 +88,10 @@ sub _init
     'width' => '1.8em',
     },
   group => { 
-    border => '1px dashed black',
+    'border' => '1px dashed black',
+    'border-style' => 'dashed',
+    'border-width' => '1',
+    'border-color' => 'black',
     background => '#a0d0ff',
     padding => '0.2em',
     },
@@ -245,11 +251,213 @@ sub node
   $self->{graph}->get_vertex_attribute( $name, OBJ );
   }
 
+#############################################################################
+# color handling
+
+my $color_names = {
+  aliceblue		=> '#f0f8ff',
+  antiquewhite		=> '#faebd7',
+  aquamarine		=> '#7fffd4',
+  aqua			=> '#00ffff',
+  azure			=> '#f0ffff',
+  beige			=> '#f5f5dc',
+  bisque		=> '#ffe4c4',
+  black			=> '#000000',
+  blanchedalmond	=> '#ffebcd',
+  blue			=> '#0000ff',
+  blueviolet		=> '#8a2be2',
+  brown			=> '#a52a2a',
+  burlywood		=> '#deb887',
+  cadetblue		=> '#5f9ea0',
+  chartreuse		=> '#7fff00',
+  chocolate		=> '#d2691e',
+  coral			=> '#ff7f50',
+  cornflowerblue	=> '#6495ed',
+  cornsilk		=> '#fff8dc',
+  crimson		=> '#dc143c',
+  cyan			=> '#00ffff',
+  darkblue		=> '#00008b',
+  darkcyan		=> '#008b8b',
+  darkgoldenrod		=> '#b8860b',
+  darkgray		=> '#a9a9a9',
+  darkgreen		=> '#006400',
+  darkgrey		=> '#a9a9a9',
+  darkkhaki		=> '#bdb76b',
+  darkmagenta		=> '#8b008b',
+  darkolivegreen	=> '#556b2f',
+  darkorange		=> '#ff8c00',
+  darkorchid		=> '#9932cc',
+  darkred		=> '#8b0000',
+  darksalmon		=> '#e9967a',
+  darkseagreen		=> '#8fbc8f',
+  darkslateblue		=> '#483d8b',
+  darkslategray		=> '#2f4f4f',
+  darkslategrey		=> '#2f4f4f',
+  darkturquoise		=> '#00ced1',
+  darkviolet		=> '#9400d3',
+  deeppink		=> '#ff1493',
+  deepskyblue		=> '#00bfff',
+  dimgray		=> '#696969',
+  dodgerblue		=> '#1e90ff',
+  firebrick		=> '#b22222',
+  floralwhite		=> '#fffaf0',
+  forestgreen		=> '#228b22',
+  fuchsia		=> '#ff00ff',
+  gainsboro		=> '#dcdcdc',
+  ghostwhite		=> '#f8f8ff',
+  goldenrod		=> '#daa520',
+  gold			=> '#ffd700',
+  gray			=> '#808080',
+  green			=> '#008000',
+  greenyellow		=> '#adff2f',
+  grey			=> '#808080',
+  honeydew		=> '#f0fff0',
+  hotpink		=> '#ff69b4',
+  indianred		=> '#cd5c5c',
+  indigo		=> '#4b0082',
+  ivory			=> '#fffff0',
+  khaki			=> '#f0e68c',
+  lavenderblush		=> '#fff0f5',
+  lavender		=> '#e6e6fa',
+  lawngreen		=> '#7cfc00',
+  lemonchiffon		=> '#fffacd',
+  lightblue		=> '#add8e6',
+  lightcoral		=> '#f08080',
+  lightcyan		=> '#e0ffff',
+  lightgoldenrodyellow	=> '#fafad2',
+  lightgray		=> '#d3d3d3',
+  lightgreen		=> '#90ee90',
+  lightgrey		=> '#d3d3d3',
+  lightpink		=> '#ffb6c1',
+  lightsalmon		=> '#ffa07a',
+  lightseagreen		=> '#20b2aa',
+  lightskyblue		=> '#87cefa',
+  lightslategray	=> '#778899',
+  lightslategrey	=> '#778899',
+  lightsteelblue	=> '#b0c4de',
+  lightyellow		=> '#ffffe0',
+  limegreen		=> '#32cd32',
+  lime			=> '#00ff00',
+  linen			=> '#faf0e6',
+  magenta		=> '#ff00ff',
+  maroon		=> '#800000',
+  mediumaquamarine	=> '#66cdaa',
+  mediumblue		=> '#0000cd',
+  mediumorchid		=> '#ba55d3',
+  mediumpurple		=> '#9370db',
+  mediumseagreen	=> '#3cb371',
+  mediumslateblue	=> '#7b68ee',
+  mediumspringgreen	=> '#00fa9a',
+  mediumturquoise	=> '#48d1cc',
+  mediumvioletred	=> '#c71585',
+  midnightblue		=> '#191970',
+  mintcream		=> '#f5fffa',
+  mistyrose		=> '#ffe4e1',
+  moccasin		=> '#ffe4b5',
+  navajowhite		=> '#ffdead',
+  navy			=> '#000080',
+  oldlace		=> '#fdf5e6',
+  olivedrab		=> '#6b8e23',
+  olive			=> '#808000',
+  orangered		=> '#ff4500',
+  orange		=> '#ffa500',
+  orchid		=> '#da70d6',
+  palegoldenrod		=> '#eee8aa',
+  palegreen		=> '#98fb98',
+  paleturquoise		=> '#afeeee',
+  palevioletred		=> '#db7093',
+  papayawhip		=> '#ffefd5',
+  peachpuff		=> '#ffdab9',
+  peru			=> '#cd853f',
+  pink			=> '#ffc0cb',
+  plum			=> '#dda0dd',
+  powderblue		=> '#b0e0e6',
+  purple		=> '#800080',
+  red			=> '#ff0000',
+  rosybrown		=> '#bc8f8f',
+  royalblue		=> '#4169e1',
+  saddlebrown		=> '#8b4513',
+  salmon		=> '#fa8072',
+  sandybrown		=> '#f4a460',
+  seagreen		=> '#2e8b57',
+  seashell		=> '#fff5ee',
+  sienna		=> '#a0522d',
+  silver		=> '#c0c0c0',
+  skyblue		=> '#87ceeb',
+  slateblue		=> '#6a5acd',
+  slategray		=> '#708090',
+  slategrey		=> '#708090',
+  snow			=> '#fffafa',
+  springgreen		=> '#00ff7f',
+  steelblue		=> '#4682b4',
+  tan			=> '#d2b48c',
+  teal			=> '#008080',
+  thistle		=> '#d8bfd8',
+  tomato		=> '#ff6347',
+  turquoise		=> '#40e0d0',
+  violet		=> '#ee82ee',
+  wheat			=> '#f5deb3',
+  white			=> '#ffffff',
+  whitesmoke		=> '#f5f5f5',
+  yellowgreen		=> '#9acd32',
+  yellow		=> '#ffff00',
+  };
+
+sub _color_as_hex
+  {
+  # Turn "red" or rgb(255,0,0) into "#ff0000". Return undef for
+  # invalid colors.
+  my ($self, $color) = @_;
+
+  $color = lc($color);
+
+  return $color_names->{$color} if exists $color_names->{$color};
+ 
+  # rgb(255,0,0) => '#ff0000' 
+  $color = sprintf("#%02x%02x%02x", $1,$2,$3)
+    if $color =~ /^rgb\(\s*(\d{1,3})\s*,\s*(\d{1,3})\s*,\s*(\d{1,3})\s*\)$/ 
+    && $1 < 256 && $2 < 256 && $3 < 256;
+
+  # turn #ff0 into #ffff00
+  $color = "#$1$1$2$2$3$3" if $color =~ /^#([a-f0-9])([a-f0-9])([a-f[0-9])\z/;
+
+  # check final color value to be #RRGGBB
+  return undef unless $color =~ /^#[a-f0-9]{6}\z/;
+
+  $color;
+  }
+
+#############################################################################
+# attribute handling
+
+sub border_attribute
+  {
+  # return "1px solid red" from the border-(style|color|width) attributes
+  my ($self, $class) = @_;
+
+  my $style = $self->attribute($class, 'border-style') || '';
+
+  return $style if $style =~ /^(none|)\z/;
+
+  my $width = $self->attribute($class, 'border-width') || '';
+  my $color = $self->attribute($class, 'border-color') || '';
+
+  $width = $width.'px' if $width =~ /^\d+\z/;
+
+  my $val = join(" ", $width, $style, $color);
+  $val =~ s/^\s+//;
+  $val =~ s/\s+\z//;
+
+  $val;
+  }
+
 sub attribute
   {
   # return the value of attribute $att from class $class
   my ($self, $class, $att) = @_;
 
+  return $self->border_attribute($class) if $att eq 'border'; # virtual attribute
+
   my $a = $self->{att};
   return undef unless exists $a->{$class} && exists $a->{$class}->{$att};
   $a->{$class}->{$att};
@@ -282,9 +490,17 @@ sub set_attribute
       }
     }
 
-  $self->{att}->{$class}->{$name} = $val;
+  if ($name eq 'border')
+    {
+    my $c = $self->{att}->{$class};
 
-  return $val;
+    ( $c->{'border-style'}, $c->{'border-width'}, $c->{'border-color'} ) = 
+	Graph::Easy::Node->border_attributes( $val ); 
+
+    return $val;
+    }
+
+  $self->{att}->{$class}->{$name} = $val;
   }
 
 sub set_attributes
@@ -344,17 +560,39 @@ sub output
   $self->$method();
   }
 
-sub css
+sub _class_styles
   {
-  my $self = shift;
+  # Create the style sheet with the class lists. This is used by both
+  # css() and As_svg(). $skip is a qr// object that returns true for
+  # attribute names to be skipped (e.g. excluded), and $map is a
+  # HASH that contains mapping for attribute names for the output (only
+  # used by As_svg()).
+  # "$base" is the basename for classes (either "table.graph$id" if 
+  # not defined, or whatever you pass in, like "" for svg).
+  # $indent is a left-indenting spacer like "  ".
+  # $overlay contains a HASH with attribute-value pairs to set as defaults.
+
+  my ($self, $skip, $map, $base, $indent, $overlay) = @_;
 
   my $a = $self->{att};
-  my $css = '';
-  my $id = $self->{id};
 
-  # for each primary class (node/group/edge) we need to find all subclasses,
-  # and list them in the CSS, too. Otherwise "node-city" would not inherit
-  # the attributes from "node".
+  $indent = '' unless defined $indent;
+
+#  my $a = $overlay;
+#  $a = {} unless defined $a;
+
+#  # Put all key/value pairs from att into overlay, overwriting possible
+#  # existing default values 
+#  foreach my $class (keys %$org_a)
+#    {
+#    my $ac = $a->{$class};
+#    foreach my $k (keys %$org_a)
+#      {
+#      $ac->{$k} = $org_a->{$k};
+#      }
+#    }
+
+  my $id = $self->{id};
 
   my $class_list = { edge => {}, node => {}, group => {} };
   foreach my $primary (qw/edge node group/)
@@ -369,17 +607,22 @@ sub css
       }
     }
 
+  $base = "table.graph$id " unless defined $base;
+
+  my $css = '';
   foreach my $class (sort keys %$a)
     {
     next if keys %{$a->{$class}} == 0;			# skip empty ones
 
     my $c = $class; $c =~ s/\./-/g;			# node.city => node-city
 
+    next if $class eq 'graph' and $base eq '';
+
     my $css_txt = '';
     my $cls = '';
     if ($class eq 'graph')
       {
-      $css_txt .= "table.graph$id {\n";
+      $css_txt .= "$indent$base\{\n";
       }
     else
       {
@@ -387,26 +630,50 @@ sub css
         {
         # generate also class list 			# like: "cities,node-rivers"
         $cls = join (",table.graph$id .$c-", sort keys %{ $class_list->{$c} });
-        $cls = ", table.graph$id .$c-$cls" if $cls ne '';		# like: ",node-cities,node-rivers"
+        $cls = ", $base.$c-$cls" if $cls ne '';		# like: ",node-cities,node-rivers"
         }
-      $css_txt .= "table.graph$id .$c$cls {\n";
+      $css_txt .= "$indent$base.$c$cls {\n";
       }
     my $done = 0;
     foreach my $att (sort keys %{$a->{$class}})
       {
-      # skip these for CSS
-      next if 
-	$att =~ /^(label|linkbase|(auto)?(link|title)|nodeclass|shape)\z/;
+      # should be skipped?
+      next if $att =~ /$skip/;
 
       $done++;						# how many did we really?
       my $val = $a->{$class}->{$att};
       # set for inner group cells "border: none"
       $val = 'none' if $att eq 'border' && $c eq 'group';
-      $css_txt .= "  $att: $val;\n";
+      $att = $map->{$att} if exists $map->{$att};	# change attribute name?
+      $css_txt .= "$indent  $att: $val;\n";
       }
-    $css_txt .= "}\n";
+    $css_txt .= "$indent}\n";
     $css .= $css_txt if $done > 0;			# skip if no attributes at all
     }
+  $css;
+  }
+
+sub _skip
+  {
+  # return the regexp that
+  my ($self) = shift;
+
+  # skip these for CSS
+  qr/^(label|linkbase|(auto)?(link|title)|nodeclass|shape)\z/;
+  }
+
+sub css
+  {
+  my $self = shift;
+
+  my $a = $self->{att};
+  my $id = $self->{id};
+
+  # for each primary class (node/group/edge) we need to find all subclasses,
+  # and list them in the CSS, too. Otherwise "node-city" would not inherit
+  # the attributes from "node".
+
+  my $css = $self->_class_styles( $self->_skip() );	
 
   # XXX TODO: we could skip this if we do not have anon nodes
   # XXX TODO: this should make anon nodes invisible, but somehow doesn't
@@ -436,14 +703,14 @@ CSS
     foreach my $group (@groups)
       {
       # could include only the ones we actually need
-      my $border = $group->attribute('border'); 
+      my $border = $group->attribute('border-style') || 'none'; 
       my $class = $group->{class}; $class =~ s/.*\.//;	# leave only subclass
       for (my $i = 1; $i <= GROUP_MAX; $i++)
 	{
         $css .= Graph::Easy::Group::Cell->_css($self->{id}, $i, $class, $border); 
 	}
       }
-    my $border = $self->attribute('group','border'); 
+    my $border = $self->attribute('group','border-style') || 'none'; 
     for (my $i = 1; $i <= GROUP_MAX; $i++)
       {
       $css .= Graph::Easy::Group::Cell->_css($self->{id}, $i, '', $border); 
@@ -887,7 +1154,7 @@ sub add_edge
 
   $self->{score} = undef;			# invalidate last layout
 
-  $self;
+  $edge;
   }
 
 sub add_node
@@ -1647,6 +1914,26 @@ inserted into the graph. In reality it should start with nodes having
 no or little incoming edges and then progress to nodes with more 
 incoming edges.
 
+A side-effect of this problem is that:
+
+	[ Bonn ] -> [ Berlin ]
+
+results in:
+
+	+------+     +--------+
+	| Bonn | --> | Berlin |
+	+------+     +--------+
+
+while this equivalent graph:
+
+	[ Berlin ] [ Bonn ] -> [ Berlin ]
+
+results in something like this:
+
+	+--------+     +------+
+	| Berlin | <-- | Bonn |
+	+--------+     +------+
+
 =head2 Grouping
 
 The output of the graphs in ASCII does not yet include the group
